"""Sports Reference team model."""

# pylint: disable=too-many-arguments,too-many-locals,duplicate-code
import datetime
import json
import urllib.parse

import extruct  # type: ignore
import pytest_is_running
import requests_cache
from bs4 import BeautifulSoup, Tag
from w3lib.html import get_base_url

from ...cache import MEMORY
from ..google.google_news_model import create_google_news_models
from ..league import League
from ..team_model import TeamModel
from ..x.x_social_model import create_x_social_model
from .sportsreference_player_model import create_sportsreference_player_model


def _create_sportsreference_team_model(
    session: requests_cache.CachedSession,
    url: str,
    dt: datetime.datetime,
    league: League,
    player_urls: set[str],
    points: float,
    fg: dict[str, int],
    fga: dict[str, int],
    offensive_rebounds: dict[str, int],
    assists: dict[str, int],
    turnovers: dict[str, int],
) -> TeamModel:
    response = session.get(url)
    response.raise_for_status()
    base_url = get_base_url(response.text, url)
    soup = BeautifulSoup(response.text, "html.parser")
    try:
        data = extruct.extract(response.text, base_url=base_url)
        name = data["json-ld"][0]["name"]
    except (json.decoder.JSONDecodeError, IndexError) as exc:
        h1 = soup.find("h1")
        if not isinstance(h1, Tag):
            raise ValueError(f"h1 is null for {url}.") from exc
        span = h1.find_all("span")
        name = span[1].get_text().strip()

    valid_player_urls = set()
    for a in soup.find_all("a"):
        player_url = urllib.parse.urljoin(url, a.get("href"))
        if player_url in player_urls:
            valid_player_urls.add(player_url)

    return TeamModel(
        identifier=name,
        name=name,
        players=[
            y
            for y in [  # pyright: ignore
                create_sportsreference_player_model(
                    session, x, dt, fg, fga, offensive_rebounds, assists, turnovers
                )
                for x in valid_player_urls
            ]
            if y is not None
        ],
        odds=[],
        points=points,
        ladder_rank=None,
        location=None,
        news=create_google_news_models(name, session, dt, league),
        social=create_x_social_model(name, session, dt),
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_sportsreference_team_model(
    session: requests_cache.CachedSession,
    url: str,
    dt: datetime.datetime,
    league: League,
    player_urls: set[str],
    points: float,
    fg: dict[str, int],
    fga: dict[str, int],
    offensive_rebounds: dict[str, int],
    assists: dict[str, int],
    turnovers: dict[str, int],
) -> TeamModel:
    return _create_sportsreference_team_model(
        session,
        url,
        dt,
        league,
        player_urls,
        points,
        fg,
        fga,
        offensive_rebounds,
        assists,
        turnovers,
    )


def create_sportsreference_team_model(
    session: requests_cache.CachedSession,
    url: str,
    dt: datetime.datetime,
    league: League,
    player_urls: set[str],
    points: float,
    fg: dict[str, int],
    fga: dict[str, int],
    offensive_rebounds: dict[str, int],
    assists: dict[str, int],
    turnovers: dict[str, int],
) -> TeamModel:
    """Create a team model from Sports Reference."""
    if not pytest_is_running.is_running():
        return _cached_create_sportsreference_team_model(
            session,
            url,
            dt,
            league,
            player_urls,
            points,
            fg,
            fga,
            offensive_rebounds,
            assists,
            turnovers,
        )
    with session.cache_disabled():
        return _create_sportsreference_team_model(
            session,
            url,
            dt,
            league,
            player_urls,
            points,
            fg,
            fga,
            offensive_rebounds,
            assists,
            turnovers,
        )



"""Sports reference league model."""

# pylint: disable=line-too-long
import datetime
import urllib.parse
from typing import Any, Iterator
from urllib.parse import parse_qs, urlparse

import requests_cache
import tqdm
from bs4 import BeautifulSoup, Tag

from ..game_model import GameModel
from ..league import League
from ..league_model import LeagueModel
from .sportsreference_game_model import create_sportsreference_game_model


class SportsReferenceLeagueModel(LeagueModel):
    """Sports Reference implementation of the league model."""

    def __init__(
        self, session: requests_cache.CachedSession, league: League, base_url: str
    ) -> None:
        super().__init__(league, session)
        self._base_url = base_url

    @property
    def games(self) -> Iterator[GameModel]:
        # pylint: disable=too-many-locals
        def _produce_games(
            div: Any, pbar: tqdm.tqdm, dt: datetime.datetime
        ) -> Iterator[GameModel]:
            for td in div.find_all("td", class_="gamelink"):
                for a in td.find_all("a"):
                    pbar.update(1)
                    game_url = urllib.parse.urljoin(url, a.get("href"))
                    game_model = create_sportsreference_game_model(
                        self.session, game_url, self.league, dt
                    )
                    pbar.set_description(
                        f"SportsReference {game_model.year} - {game_model.season_type} - {game_model.dt}"
                    )
                    yield game_model

        final_path: str | None = ""
        with tqdm.tqdm() as pbar:
            while final_path is not None:
                url = self._base_url + final_path
                if final_path:
                    response = self.session.get(url)
                else:
                    with self.session.cache_disabled():
                        response = self.session.get(url)
                response.raise_for_status()
                dt = datetime.datetime.now()
                if final_path:
                    parsed_url = urlparse(url)
                    query = parse_qs(parsed_url.query)
                    dt = datetime.datetime(
                        int(query["year"][0]),
                        int(query["month"][0]),
                        int(query["day"][0]),
                    )
                soup = BeautifulSoup(response.text, "html.parser")
                if self.league == League.NCAAB:
                    for div in soup.find_all("div", class_="gender-m"):
                        yield from _produce_games(div, pbar, dt)
                else:
                    yield from _produce_games(soup, pbar, dt)
                prev_a = soup.find("a", class_="prev")
                if isinstance(prev_a, Tag):
                    href = prev_a.get("href")
                    if isinstance(href, str):
                        prev_url = urllib.parse.urljoin(url, href)
                        final_path = prev_url.split("/")[-1]
                    else:
                        final_path = None
                else:
                    final_path = None



"""Odds Portal team model."""

# pylint: disable=too-many-arguments,duplicate-code,line-too-long,too-many-locals
import base64
import datetime
import json

import pytest_is_running
import requests_cache
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from ...cache import MEMORY
from ..google.google_news_model import create_google_news_models
from ..league import League
from ..team_model import TeamModel
from ..x.x_social_model import create_x_social_model
from .oddsportal_odds_model import create_oddsportal_odds_model


def _create_oddsportal_team_model(
    session: requests_cache.CachedSession,
    dt: datetime.datetime,
    team_name: str,
    league: League,
    points: float,
    version_id: str,
    sport_id: str,
    unique_id: str,
    default_bet_id: str,
    default_scope_id: str,
    xhash: str,
    salt: bytes,
    password: bytes,
    bookie_names: dict[str, str],
    team_idx: int,
) -> TeamModel:
    response = session.get(
        f"https://www.oddsportal.com/feed/match-event/{version_id}-{sport_id}-{unique_id}-{default_bet_id}-{default_scope_id}-{xhash}.dat"
    )
    response.raise_for_status()
    decoded_data = base64.b64decode(response.content).decode()
    encrypted, key = decoded_data.split(":")
    encrypted_bytes = base64.urlsafe_b64decode(encrypted)
    key_bytes = bytes.fromhex(key)
    kdf = PBKDF2HMAC(
        algorithm=SHA256(),
        length=32,
        salt=salt,
        iterations=1000,
        backend=default_backend(),
    )
    aes_key = kdf.derive(password)
    cipher = Cipher(
        algorithms.AES(aes_key), modes.CBC(key_bytes), backend=default_backend()
    )
    decryptor = cipher.decryptor()
    decrypted_bytes = decryptor.update(encrypted_bytes) + decryptor.finalize()
    decrypted_data = decrypted_bytes.decode("utf-8")
    end_of_json = decrypted_data.rfind("}")
    if end_of_json != -1:
        decrypted_data = decrypted_data[: end_of_json + 1]
    parsed_data = json.loads(decrypted_data)
    odds_data = parsed_data["d"]["oddsdata"]["back"][
        f"E-{default_bet_id}-{default_scope_id}-0-0-0"
    ]
    outcome_id = odds_data["outcomeId"][team_idx]
    history = odds_data["history"][outcome_id]
    odds_models = []
    for bookie_id, bookie_name in bookie_names.items():
        for odds, _, timestamp in history[bookie_id]:
            odds_models.append(
                create_oddsportal_odds_model(
                    odds,
                    datetime.datetime.fromtimestamp(timestamp),
                    bookie_name,
                    bookie_id,  # type: ignore
                )
            )

    return TeamModel(
        identifier=team_name,
        name=team_name,
        points=points,
        players=[],
        odds=odds_models,
        ladder_rank=None,
        location=None,
        news=create_google_news_models(team_name, session, dt, league),
        social=create_x_social_model(team_name, session, dt),
        field_goals=None,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_oddsportal_team_model(
    session: requests_cache.CachedSession,
    dt: datetime.datetime,
    team_name: str,
    league: League,
    points: float,
    version_id: str,
    sport_id: str,
    unique_id: str,
    default_bet_id: str,
    default_scope_id: str,
    xhash: str,
    salt: bytes,
    password: bytes,
    bookie_names: dict[str, str],
    team_idx: int,
) -> TeamModel:
    return _create_oddsportal_team_model(
        session,
        dt,
        team_name,
        league,
        points,
        version_id,
        sport_id,
        unique_id,
        default_bet_id,
        default_scope_id,
        xhash,
        salt,
        password,
        bookie_names,
        team_idx,
    )


def create_oddsportal_team_model(
    session: requests_cache.CachedSession,
    dt: datetime.datetime,
    team_name: str,
    league: League,
    points: float,
    version_id: str,
    sport_id: str,
    unique_id: str,
    default_bet_id: str,
    default_scope_id: str,
    xhash: str,
    salt: bytes,
    password: bytes,
    bookie_names: dict[str, str],
    team_idx: int,
) -> TeamModel:
    """Create a team model based off the odds portal response."""
    if not pytest_is_running.is_running() and dt < datetime.datetime.now().replace(
        tzinfo=dt.tzinfo
    ) - datetime.timedelta(days=7):
        return _cached_create_oddsportal_team_model(
            session,
            dt,
            team_name,
            league,
            points,
            version_id,
            sport_id,
            unique_id,
            default_bet_id,
            default_scope_id,
            xhash,
            salt,
            password,
            bookie_names,
            team_idx,
        )
    with session.cache_disabled():
        return _create_oddsportal_team_model(
            session,
            dt,
            team_name,
            league,
            points,
            version_id,
            sport_id,
            unique_id,
            default_bet_id,
            default_scope_id,
            xhash,
            salt,
            password,
            bookie_names,
            team_idx,
        )



"""NFL aussportsbetting league model."""

import requests_cache

from ...aussportsbetting.aussportsbetting_league_model import \
    AusSportsBettingLeagueModel
from ...league import League


class NFLAusSportsBettingLeagueModel(AusSportsBettingLeagueModel):
    """NFL AusSportsBetting implementation of the league model."""

    def __init__(self, session: requests_cache.CachedSession) -> None:
        super().__init__(League.NFL, session)



"""Odds Portal league model."""

import json
import urllib.parse
from typing import Iterator

import extruct  # type: ignore
import tqdm
from bs4 import BeautifulSoup, Tag

from ..game_model import GameModel
from ..league import League
from ..league_model import LeagueModel
from .oddsportal_game_model import create_oddsportal_game_model


class OddsPortalLeagueModel(LeagueModel):
    """Odds Portal implementation of the league model."""

    @property
    def _path(self) -> str:
        match self.league:
            case League.AFL:
                return "/".join(["aussie-rules", "australia", "afl", ""])
            case League.NBA:
                return "/".join(["basketball", "usa", "nba", ""])
            case League.NCAAB:
                return "/".join(["basketball", "usa", "ncaa", ""])
            case League.NFL:
                return "/".join(["american-football", "usa", "nfl", ""])
            case _:
                raise ValueError(f"Unsupported league: {self.league}")

    def _find_next(self, pbar: tqdm.tqdm) -> Iterator[GameModel]:
        base_url = "https://www.oddsportal.com/" + self._path
        with self.session.cache_disabled():
            response = self.session.get(base_url)
        response.raise_for_status()
        data = extruct.extract(response.text, base_url=base_url)
        for jsonld in data["json-ld"]:
            if jsonld["@type"] != "SportsEvent":
                continue
            game_model = create_oddsportal_game_model(
                self.session, jsonld["url"], self.league, True
            )
            pbar.update(1)
            pbar.set_description(f"OddsPortal {game_model.dt}")
            yield game_model

    def _find_previous(self, pbar: tqdm.tqdm) -> Iterator[GameModel]:
        seen_urls = set()
        queued_urls = {"https://www.oddsportal.com/" + self._path + "results/"}
        while queued_urls:
            url = queued_urls.pop()
            if url in seen_urls:
                continue
            seen_urls.add(url)

            response = self.session.get(url)
            response.raise_for_status()

            soup = BeautifulSoup(response.text, "html.parser")

            # Find next URLs
            for a in soup.find_all("a", href=True):
                next_url = urllib.parse.urljoin(url, a.get("href"))
                if next_url.endswith("/results/"):
                    queued_urls.add(next_url)

            # Paginate through results
            while True:
                tournament_component = soup.find("tournament-component")
                if not isinstance(tournament_component, Tag):
                    raise ValueError("tournament_component is not a tag.")
                matches = json.loads(str(tournament_component[":sport-data"]))
                for match in matches["tournamentGamesComponent"]["rows"]:
                    game_model = create_oddsportal_game_model(
                        self.session, match["url"], self.league, False
                    )
                    pbar.update(1)
                    pbar.set_description(f"OddsPortal {game_model.dt}")
                    yield game_model

                next_a = soup.find("a", text="Next", href=True)
                if next_a is None:
                    break
                if not isinstance(next_a, Tag):
                    raise ValueError("next_a is not a tag.")
                next_url = urllib.parse.urljoin(url, str(next_a.get("href")))
                response = self.session.get(next_url)
                response.raise_for_status()
                soup = BeautifulSoup(response.text, "html.parser")

    @property
    def games(self) -> Iterator[GameModel]:
        with tqdm.tqdm() as pbar:
            yield from self._find_next(pbar)
            yield from self._find_previous(pbar)



"""Tests for the NBA NBA player model class."""
import datetime
import unittest

import requests_mock
import requests_cache
from sportsball.data.nba.nba.nba_nba_player_model import create_nba_nba_player_model
import pandas as pd


class TestNBANBAPlayerModel(unittest.TestCase):

    def setUp(self):
        self.session = requests_cache.CachedSession(backend="memory")

    def test_identifier(self):
        dt = datetime.datetime(2010, 10, 10, 10, 10, 00)
        with requests_mock.Mocker() as m:
            m.get("https://stats.nba.com/stats/commonplayerinfo?LeagueID=&PlayerID=a")
            player_model = create_nba_nba_player_model(
                pd.Series(data=[
                    "a",
                ], index=[
                    "PERSON_ID",
                ]),
                pd.DataFrame(index=[], columns=["PERSON_ID"]),
                dt,
                self.session,
            )
            self.assertEqual(player_model.identifier, "a")
